// ============================================================================
// Module Name: IPat_DcFifo
// Revision: 20170930a
// Author: Altmo toolbox
// Description:
//  + Dual clock asynchronous FIFO
// ============================================================================

`timescale 1ps/100fs

module IPat_DcFifo (
    wrclk,	// write clock
    data,	// write data
    wrreq,	// write request
    wrfull,	// write full
    wrusedw,	// write used word
    
    rdclk,	// read clock
    rdreq,	// read request
    q,		// read data
    rdempty,	// read empty
    rdusedw,	// read used wordw
    
    rst_n	// asynchronous reset 0 active
);

// parameter ==================================================================
parameter P_IPAT_DCFIFO_DWIDTH = 8;		// data width
parameter P_IPAT_DCFIFO_AWIDTH = 4;		// address width
parameter P_IPAT_DCFIFO_WRPIPE = 2;		// write sync pipe stage
parameter P_IPAT_DCFIFO_RDPIPE = 2;		// read sync pipe stage
parameter P_IPAT_DCFIFO_SAHEAD = "FALSE";	// read mode norma/show-ahead
parameter P_IPAT_DCFIFO_OUTFF  = "FALSE";	// output ff enable

parameter P_IPAT_DCFIFO_DBGMSG = "FALSE";	// debug message

// local parameter ------------------------------------------------------------
localparam LP_DWIDTH = P_IPAT_DCFIFO_DWIDTH;
localparam LP_AWIDTH = P_IPAT_DCFIFO_AWIDTH;
localparam LP_WRPIPE = P_IPAT_DCFIFO_WRPIPE;
localparam LP_RDPIPE = P_IPAT_DCFIFO_RDPIPE;
localparam LP_SAHEAD = P_IPAT_DCFIFO_SAHEAD;
localparam LP_OUTFF  = P_IPAT_DCFIFO_OUTFF;
localparam LP_DBGMSG = P_IPAT_DCFIFO_DBGMSG;

// synopsys translate_off
initial begin
    if (LP_DBGMSG == "TRUE") begin
        $display("INFO:%m:LP_DWIDTH=%0d",LP_DWIDTH);
        $display("INFO:%m:LP_AWIDTH=%0d",LP_AWIDTH);
        $display("INFO:%m:LP_WRPIPE=%0d",LP_WRPIPE);
        $display("INFO:%m:LP_RDPIPE=%0d",LP_RDPIPE);
        $display("INFO:%m:LP_SAHEAD=%s", LP_SAHEAD);
        $display("INFO:%m:LP_OUTFF=%s"  ,LP_OUTFF);
    end
end
// synopsys translate_on


// port definition ============================================================
input			wrclk;		// write clock
input	[LP_DWIDTH-1:0]	data;		// write data
input			wrreq;		// write request
output			wrfull;		// write full
output	[LP_AWIDTH:0]	wrusedw;	// write used word

input			rdclk;		// read clock
input			rdreq;		// read request
output	[LP_DWIDTH-1:0]	q;		// read data
output			rdempty;	// read empty
output	[LP_AWIDTH:0]	rdusedw;	// read used word

input			rst_n;		// asynchronous reset 0 active

// wire drive-module_signal ----------------------------------------------------
wire			SynRstWr_rst_n_sync;	// synchronized reset for wrclk
wire			SynRstRd_rst_n_sync;	// synchronized reset for rdclk

wire	[LP_AWIDTH:0]	Gray2BinRcadr_o_bin;	// read address(binary) with carry bit
wire	[LP_AWIDTH:0]	Gray2BinWcadr_o_bin;	// write address(binary) with carry bit

wire	[LP_AWIDTH:0]	Bin2GrayRcadr_o_gray;	// read address(gray) with carry bit
wire	[LP_AWIDTH:0]	Bin2GrayWcadr_o_gray;	// write address(gray) with carry bit

wire	[LP_AWIDTH:0]	SyncVarRcadr_o_sync;	// read address(gray) with carry bit sync with wrclk
wire	[LP_AWIDTH:0]	SyncVarWcadr_o_sync;	// write address(gray) with carry bit sync with rdclk

wire	[LP_AWIDTH-1:0]	DcFifoWrCtrl_o_wradr;	// write address
wire			DcFifoWrCtrl_o_wren;	// write enable
wire	[LP_AWIDTH:0]	DcFifoWrCtrl_o_wrcadr;	// write address with carry bit

wire	[LP_AWIDTH-1:0]	DcFifoRdCtrl_o_rdadr;	// read address
wire			DcFifoRdCtrl_o_rden;	// read enable
wire	[LP_AWIDTH:0]	DcFifoRdCtrl_o_rdcadr;	// read address with carry bit

wire	[LP_DWIDTH-1:0]	RamSdpWrap_q;		// read data from ram


// simple dual port ram =======================================================
// If you'd like to use block ram of FPGA or ASIC,
// you can change following description of instance to use your RAM module.
// ----------------------------------------------------------------------------
IPat_RamSdpWrap #(
	.P_IPAT_RAMSDPWRAP_DWIDTH(LP_DWIDTH),	// data width
	.P_IPAT_RAMSDRWRAP_AWIDTH(LP_AWIDTH)	// address width
) RamSdpWrap (
	.wrclock(wrclk),
	.data(data),
	.wraddress(DcFifoWrCtrl_o_wradr),
	.wren(DcFifoWrCtrl_o_wren),
	.rdclock(rdclk),
	.rdaddress(DcFifoRdCtrl_o_rdadr),
	.rden(DcFifoRdCtrl_o_rden),
	.q(RamSdpWrap_q)
);
// ============================================================================


// graycode converter for write address
IPat_Bin2Gray #(
	.P_IPAT_BIN2GRAY_DWIDTH(LP_AWIDTH+1'b1)
) Bin2GrayWcadr (
	.i_bin(DcFifoWrCtrl_o_wrcadr),
	.o_gray(Bin2GrayWcadr_o_gray)
);

// graycode converter for read address
IPat_Bin2Gray #(
	.P_IPAT_BIN2GRAY_DWIDTH(LP_AWIDTH+1'b1)
) Bin2GrayRcadr (
	.i_bin(DcFifoRdCtrl_o_rdcadr),
	.o_gray(Bin2GrayRcadr_o_gray)
);

// gray code synchronizer for read clock
IPat_SyncVar #(
	.P_IPAT_SYNCVAR_DWIDTH(LP_AWIDTH+1'b1),	// address width with carry bit
	.P_IPAT_SYNCVAR_PDEPTH(LP_WRPIPE)	// depth of pipe for write clock
) SyncVarWcadr (
	.clk(rdclk),
	.i_sync(Bin2GrayWcadr_o_gray),
	.o_sync(SyncVarWcadr_o_sync),
	.rst_n(SynRstRd_rst_n_sync)
);

// gray code synchronizer for write clock
IPat_SyncVar #(
	.P_IPAT_SYNCVAR_DWIDTH(LP_AWIDTH+1'b1),	// address width with carry bit
	.P_IPAT_SYNCVAR_PDEPTH(LP_WRPIPE)	// depth of pipe for write clock
) SyncVarRcadr (
	.clk(wrclk),
	.i_sync(Bin2GrayRcadr_o_gray),
	.o_sync(SyncVarRcadr_o_sync),
	.rst_n(SynRstWr_rst_n_sync)
);

// binary converter for write address
IPat_Gray2Bin #(
	.P_IPAT_GRAY2BIN_DWIDTH(LP_AWIDTH+1'b1)	// address width with carry bit
) Gray2BinWcadr (
	.i_gray(SyncVarWcadr_o_sync),
	.o_bin(Gray2BinWcadr_o_bin)
);

// binary converter for read address
IPat_Gray2Bin #(
	.P_IPAT_GRAY2BIN_DWIDTH(LP_AWIDTH+1'b1)	// address width with carry bit
) Gray2BinRcadr (
	.i_gray(SyncVarRcadr_o_sync),
	.o_bin(Gray2BinRcadr_o_bin)
);

// read control
IPat_DcFifoRdCtrl #(
	.P_IPAT_DCFIFORDCTRL_AWIDTH(LP_AWIDTH),	// address width
	.P_IPAT_DCFIFORDCTRL_SAHEAD(LP_SAHEAD),	// read mode
	.P_IPAT_DCFIFORDCTRL_DBGMSG(LP_DBGMSG)	// debug message
) DcFifoRdCtrl (
	.clk(rdclk),
	.rst_n(SynRstRd_rst_n_sync),
	.i_rdreq(rdreq),
	.i_wrcadr(Gray2BinWcadr_o_bin),
	.o_rdcadr(DcFifoRdCtrl_o_rdcadr),
	.o_rdadr(DcFifoRdCtrl_o_rdadr),
	.o_rden(DcFifoRdCtrl_o_rden),
	.o_rdempty(rdempty),
	.o_rdusedw(rdusedw)
);

// write control
IPat_DcFifoWrCtrl #(
	.P_IPAT_DCFIFOWRCTRL_AWIDTH(LP_AWIDTH),	// address width
	.P_IPAT_DCFIFOWRCTRL_DBGMSG(LP_DBGMSG)	// debug message
) DcFifoWrCtrl (
	.clk(wrclk),
	.rst_n(SynRstWr_rst_n_sync),
	.i_wrreq(wrreq),
	.i_rdcadr(Gray2BinRcadr_o_bin),
	.o_wrcadr(DcFifoWrCtrl_o_wrcadr),
	.o_wradr(DcFifoWrCtrl_o_wradr),
	.o_wren(DcFifoWrCtrl_o_wren),
	.o_wrfull(wrfull),
	.o_wrusedw(wrusedw)
);

// output clear logic
IPat_FifoOutRst #(
	.P_IPAT_FIFOOUTRST_DWIDTH(LP_DWIDTH),	// data width
	.P_IPAT_FIFOOUTRST_OUTFF(LP_OUTFF)	// output ff enable
) FifoOutRst (
	.clk(rdclk),
	.i_q(RamSdpWrap_q),
	.o_q(q),
	.i_rden(DcFifoRdCtrl_o_rden),
	.rst_n(SynRstRd_rst_n_sync)
);

// reset synchronizer for wrclk
IPat_SyncRst SynRstWr (
	.clk(wrclk),
	.rst_n(rst_n),
	.rst_n_sync(SynRstWr_rst_n_sync)
);

// reset synchronizer for rdclk
IPat_SyncRst SynRstRd (
	.clk(rdclk),
	.rst_n(rst_n),
	.rst_n_sync(SynRstRd_rst_n_sync)
);

endmodule
