// ============================================================================
// Module Name: IPat_DcFifoRdCtrl
// Revision: 20170930a
// Author: Altmo toolbox
// Description:
//  + Read adress control of IPat_DcFifo
// ============================================================================

`timescale 1ps/100fs

module IPat_DcFifoRdCtrl (
    clk,
    rst_n,
    i_rdreq,
    i_wrcadr,
    o_rdadr,
    o_rdcadr,
    o_rden,
    o_rdempty,
    o_rdusedw
);

// parameter ==================================================================
parameter P_IPAT_DCFIFORDCTRL_AWIDTH = 4;	// address width
parameter P_IPAT_DCFIFORDCTRL_SAHEAD = "FALSE";	// Show-ahead ON/OFF
parameter P_IPAT_DCFIFORDCTRL_DBGMSG = "FALSE";	// debug message

// local parameter ============================================================
localparam LP_AWIDTH  = P_IPAT_DCFIFORDCTRL_AWIDTH;	// address width
localparam LP_SAHEAD  = P_IPAT_DCFIFORDCTRL_SAHEAD;	// show-ahead
localparam LP_DBGMSG  = P_IPAT_DCFIFORDCTRL_DBGMSG;	// debug message

localparam LP_UD = 1;	// unit delay to avoid racing

// synopsys translate_off
initial begin
    if (LP_DBGMSG == "TRUE") begin
        $display("INFO_IPat:%m:LP_AWIDTH=%0d",LP_AWIDTH);
        $display("INFO_IPat:%m:LP_SAHEAD=%s" ,LP_SAHEAD);
    end
end
// synopsys translate_on


// port definition ============================================================
input			clk;		// clock
input			rst_n;		// reset 0 active
input			i_rdreq;	// read request
input	[LP_AWIDTH:0]	i_wrcadr;	// carry + write address

output	[LP_AWIDTH-1:0]	o_rdadr;	// read address
output	[LP_AWIDTH:0]	o_rdcadr;	// carry + read address
output			o_rden;		// read enable to memory
output			o_rdempty;	// read empty signal
output	[LP_AWIDTH:0]	o_rdusedw;	// used write word

// wire definition ------------------------------------------------------------
wire			w_rdempty;
wire	[LP_AWIDTH:0]	w_rdusedw;

// carry bit + read address ---------------------------------------------------
reg	[LP_AWIDTH:0]	r_rdcadr;

always @(posedge clk or negedge rst_n) begin
    if (rst_n == 1'b0) begin
        r_rdcadr <= #LP_UD {(LP_AWIDTH+1){1'b0}};
    end
    else begin
        r_rdcadr <= #LP_UD (o_rden)? (r_rdcadr + 1'b1) : r_rdcadr;
    end
end

assign o_rdadr  = r_rdcadr[LP_AWIDTH-1:0]; // remove carry bit
assign o_rdcadr = r_rdcadr;

// read empty and read enable -------------------------------------------------
assign w_rdempty = (i_wrcadr == r_rdcadr); // same address with carry

generate
    if (LP_SAHEAD=="TRUE") begin // show-ahead mode
        reg	r_rdempty;
        wire	w_showahead;
        
        always @(posedge clk or negedge rst_n) begin
            if (rst_n == 1'b0) begin
                r_rdempty <= #LP_UD 1'b1;
            end
            else begin
                r_rdempty <= #LP_UD w_rdempty;
            end
        end
        
        assign w_showahead = r_rdempty & ~w_rdempty;   // show-ahead condition
        
        assign o_rden    = (~w_rdempty & ~r_rdempty & i_rdreq) |    // normal mode
                           w_showahead;                // show-ahead mode
        
        assign o_rdempty =   r_rdempty; // o_rdempty comes from output of flip-flop
    end
    else begin // normal mode
        assign o_rdempty =  w_rdempty;            // no flip-flop
        assign o_rden    = ~w_rdempty & i_rdreq;  // read request controls read enable
    end
endgenerate

// read used word -------------------------------------------------------------
assign w_rdusedw = i_wrcadr - r_rdcadr;

generate
    if (LP_SAHEAD=="TRUE") begin // show-ahead mode
        reg	[LP_AWIDTH:0]	r_rdusedw;
        
        always @(posedge clk or negedge rst_n) begin
            if (rst_n == 1'b0) begin
                r_rdusedw <= #LP_UD {(LP_AWIDTH+1){1'b0}};
            end
            else begin
                r_rdusedw <= #LP_UD w_rdusedw;
            end
        end
        
        assign o_rdusedw = r_rdusedw;
    end
    else begin
        assign o_rdusedw = w_rdusedw;
    end
endgenerate

endmodule
