// ============================================================================
// Module Name: IPat_RamSdpWrap
// Revision: 20170923a
// Author: Altmo toolbox
// Description:
//  + Behavior of IPat_RamSdpWrap is simple dual port RAM.
//  + User can synthesis this module.
//  + User can replace this module to FPGA or ASIC RAM module for reducing area
//    and power consumption.
//  + This module doesn't support RDW(read during write in same address).
// ============================================================================

`timescale 1ps/100fs

module IPat_RamSdpWrap (
    wrclock,
    data,
    wraddress,
    wren,
    rdclock,
    q,
    rdaddress,
    rden
);

// parameter ==================================================================
parameter P_IPAT_RAMSDPWRAP_DWIDTH = 8;	// width of data and q
parameter P_IPAT_RAMSDRWRAP_AWIDTH = 4;	// width of wraddress and rdaddress

// local parameter ============================================================
localparam LP_DWIDTH = P_IPAT_RAMSDPWRAP_DWIDTH;
localparam LP_AWIDTH = P_IPAT_RAMSDRWRAP_AWIDTH;
localparam LP_ADEPTH = (1 << LP_AWIDTH);	// word depth

localparam LP_UD = 1;	// unit delay to avoid racing

// port definition ============================================================
input			wrclock;	// write clock
input	[LP_DWIDTH-1:0]	data;		// write data
input	[LP_AWIDTH-1:0]	wraddress;	// write address
input			wren;		// write enable

input			rdclock;	// read clock
input	[LP_AWIDTH-1:0]	rdaddress;	// read address
input			rden;		// read enable

output	[LP_DWIDTH-1:0]	q;		// read data

// register and wire definition ===============================================
reg	[LP_DWIDTH-1:0]	mem[0:LP_ADEPTH-1];	// memory register array
reg	[LP_AWIDTH-1:0]	r_rdaddress;	// read address register

// write ----------------------------------------------------------------------
always @(posedge wrclock) begin
    if (wren) begin
        mem[wraddress] <= #LP_UD data;
    end
end

// read -----------------------------------------------------------------------
always @(posedge rdclock) begin
    if (rden) begin
        r_rdaddress <= #LP_UD rdaddress;
    end
end

assign q = mem[r_rdaddress];

// debug ----------------------------------------------------------------------
// synopsys translate_off
reg	[LP_AWIDTH-1:0]	r_wraddress;	// write address register
reg			r_wren;		// write enable register
reg			r_rden;		// read enable register

always @(posedge wrclock) begin
    r_wraddress <= #LP_UD (wren)? wraddress : r_wraddress;
    r_wren <= #LP_UD wren;
end

always @(posedge rdclock) begin
    r_rden <= #LP_UD rden;
end

wire	w_detect_rdw;	// RDW detection
assign w_detect_rdw = ((r_wraddress == r_rdaddress) & (r_wren & r_rden));

always @(posedge w_detect_rdw) begin
    $display("WARNING: Read during write at %t in %m",$stime);
end
// synopsys translate_on

endmodule
